/*
 * This Java source file was generated by the Gradle 'init' task.
 */
package HpdosClient;

import HpdosClient.MessageFormat.MessageConstants;
import HpdosClient.MessageFormat.RequestBuilder;
import com.google.common.base.Stopwatch;
import hpdos.grpc.NetworkServiceGrpc;
import hpdos.grpc.Packet;
import hpdos.grpc.PacketFormatProto;
import hpdos.grpc.Request;
import io.grpc.ManagedChannel;
import io.grpc.ManagedChannelBuilder;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.sql.Timestamp;
import java.time.Instant;
import java.util.*;
import java.util.concurrent.*;

public class ClientRunner {
    public static int parallelCount = 10;
    public static int runtime = 0;
    private final String clientID;
    public static String propertiesFile;
    private int cCreate, cRead, cUpdate, cDelete;

    private Queue<Long> createTime, updateTime, readTime, deleteTime;

    public ClientRunner() {
        clientID = UUID.randomUUID().toString();
        Properties properties = new Properties();
        try {
            InputStream inputStream = new FileInputStream(propertiesFile);
            properties.load(inputStream);
            parallelCount = Integer.parseInt((String) properties.get("app.thread_count"));
            runtime = Integer.parseInt((String) properties.get("app.runtime"));
            cCreate = Integer.parseInt((String) properties.get("app.cycle_create"));
            cRead = Integer.parseInt((String) properties.get("app.cycle_read"));
            cUpdate = Integer.parseInt((String) properties.get("app.cycle_update"));
            cDelete = Integer.parseInt((String) properties.get("app.cycle_delete"));

            createTime = new ConcurrentLinkedQueue<>();
            updateTime = new ConcurrentLinkedQueue<>();
            readTime = new ConcurrentLinkedQueue<>();
            deleteTime = new ConcurrentLinkedQueue<>();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }
    public String getGreeting() {
        return "Hello World!";
    }

    // create a metadata block
    public long create(NetworkServiceGrpc.NetworkServiceBlockingStub stub, String key, String value) {
        ArrayList<Request> request = new ArrayList<>();
        request.add(RequestBuilder.buildRequest(MessageConstants.METADATA_CREATE,
                0, value.length(), key,
                0, MessageConstants.METADATA_ACCESS_PRIVATE, this.clientID, value));
        Packet packet = RequestBuilder.buildPacket(request);
        long timestampCreateStart = System.currentTimeMillis();
        Packet response = stub.createMetadata(packet);
        return timestampCreateStart;
    }

    // read back the metadata
    public Map.Entry<Packet, Long> read(NetworkServiceGrpc.NetworkServiceBlockingStub stub, String key) {
        ArrayList<Request> request = new ArrayList<>();
        request.add(RequestBuilder.buildRequest(MessageConstants.METADATA_READ,
                0, 0, key, 0, MessageConstants.METADATA_ACCESS_PRIVATE, this.clientID, ""));
        Packet packet = RequestBuilder.buildPacket(request);
        long timestampReadStart = System.currentTimeMillis();
        Packet response = stub.readMetadata(packet);
//        System.out.println(response);
        return new AbstractMap.SimpleEntry<>(response, timestampReadStart);
    }

    public long update(NetworkServiceGrpc.NetworkServiceBlockingStub stub, String key, String value, int version) {
        ArrayList<Request> request = new ArrayList<>();
        request.add(RequestBuilder.buildRequest(MessageConstants.METADATA_UPDATE,
                version, value.length(), key,
                0, MessageConstants.METADATA_ACCESS_PRIVATE, this.clientID, value));
        Packet packet = RequestBuilder.buildPacket(request);
        long timestampCreateStart = System.currentTimeMillis();
        Packet response = stub.updateMetadata(packet);
        return timestampCreateStart;
    }

    public long delete(NetworkServiceGrpc.NetworkServiceBlockingStub stub, String key, int version) {
        ArrayList<Request> request = new ArrayList<>();
        request.add(RequestBuilder.buildRequest(MessageConstants.METADATA_DELETE,
                version, 0, key,
                0, MessageConstants.METADATA_ACCESS_PRIVATE, this.clientID, ""));
        Packet packet = RequestBuilder.buildPacket(request);
        long timestampCreateStart = System.currentTimeMillis();
        Packet response = stub.deleteMetadata(packet);
        return timestampCreateStart;
    }

    public double initClient(String id, long experimentStartTime) {
        final ManagedChannel channel = ManagedChannelBuilder.
                forAddress(ConfigConstants.HOST, ConfigConstants.PORT)
                .usePlaintext()
                .build();

        long startTime = System.currentTimeMillis();
        double i = 0;
        for (;;i++) {

            String key = id + (int) (Math.random() * Integer.MAX_VALUE),
                    value = "dummy",
                    updatedValue = "dummyUpdated";
            NetworkServiceGrpc.NetworkServiceBlockingStub stub = NetworkServiceGrpc.newBlockingStub(channel);
            for (int j = 0; j < cCreate; j++) {
                long timestampCreateStart = create(stub, key, value);
                createTime.add(System.currentTimeMillis() - timestampCreateStart);
            }
            for (int j = 0; j < cRead; j++) {
                AbstractMap.Entry<Packet, Long> data = read(stub, key);
                readTime.add(System.currentTimeMillis() - data.getValue());
            }

            for (int j = 0; j < cUpdate; j++) {
                AbstractMap.Entry<Packet, Long> data = read(stub, key);
                long timestampUpdateStart = update(stub, key, updatedValue,
                        data.getKey().getResponse(0).getAck().getVersion());
                readTime.add(System.currentTimeMillis() - data.getValue());
                updateTime.add(System.currentTimeMillis() - timestampUpdateStart);
            }

            for (int j = 0; j < cDelete; j++) {
                AbstractMap.Entry<Packet, Long> data = read(stub, key);
                long timestampDeleteStart = delete(stub, key, data.getKey().getResponse(0).getAck().getVersion());
                readTime.add(System.currentTimeMillis() - data.getValue());
                deleteTime.add(System.currentTimeMillis() - timestampDeleteStart);
            }
            long currentTime = System.currentTimeMillis();
            if ((currentTime - experimentStartTime) >= runtime * 1000L)
                break;
        }
//        System.out.println(id + "runtime " + (System.currentTimeMillis() - startTime) +
//                "ms qps " + qps);
        channel.shutdown();
        return 0;
    }

    public static void main(String[] args) throws InterruptedException, ExecutionException {
        propertiesFile = args[0];
        ClientRunner clientRunner = new ClientRunner();
        System.out.println(clientRunner.getGreeting());
        System.out.println("Thread count: " + parallelCount + " runtime: " + runtime + "s");
        Thread.sleep(1000);
        ExecutorService executorService = Executors.newFixedThreadPool(parallelCount);
        Set<Callable<Double>> callables = new HashSet<>();
        final long startTime = System.currentTimeMillis();
        for (int i = 0; i < parallelCount; i++) {
            int finalI = i;
            callables.add(() -> clientRunner.initClient(Integer.toString(finalI), startTime));
        }

        ScheduledExecutorService scheduledExecutorService = Executors.newSingleThreadScheduledExecutor() ;

        Stopwatch stopwatch = Stopwatch.createUnstarted();
        Runnable task = () -> {
            System.out.println( "Experiment running: " + stopwatch );
        };
        scheduledExecutorService.schedule( task, 5, TimeUnit.SECONDS) ;
        stopwatch.start();
        List<Future<Double>> futures = executorService.invokeAll(callables);
        double qps = 0;
        for (Future<Double> future: futures) {
            future.get();
        }
        long endTime = System.currentTimeMillis();
        double totalRuntime = endTime - startTime;
        long readQps = 0, createQps = 0, updateQps = 0, deleteQps = 0;
        double avgRead = 0, avgCreate = 0, avgUpdate = 0, avgDelete = 0;
        for (Long time: clientRunner.readTime) {
            readQps++;
            avgRead += time;
        }
        avgRead /= readQps * 1.0;
        for (Long time: clientRunner.createTime) {
            createQps++;
            avgCreate += time;
        }
        avgCreate /= createQps * 1.0;
        for (Long time: clientRunner.updateTime) {
            updateQps++;
            avgUpdate += time;
        }
        avgUpdate /= updateQps * 1.0;
        for (Long time: clientRunner.deleteTime) {
            deleteQps++;
            avgDelete += time;
        }
        avgDelete /= deleteQps * 1.0;
        double totalQps = readQps + createQps + updateQps + deleteQps;
        System.out.println("Total runtime: " + totalRuntime);
        System.out.println("QPS Read: " + readQps + " Create: " + createQps + " Update: " + updateQps + " Delete: " + deleteQps);
        System.out.println("QPS Total: " + totalQps + " avg query time[thread amortized]: " + (totalQps / (totalRuntime * parallelCount)));
        System.out.println("Read Total: " + readQps + " avg query time: " + avgRead);
        System.out.println("Create Total: " + createQps + " avg query time: " + avgCreate);
        System.out.println("Update Total: " + updateQps + " avg query time: " + avgUpdate);
        System.out.println("Delete Total: " + deleteQps + " avg query time: " + avgDelete);
//        clientRunner.initClient();
        executorService.shutdown();
        scheduledExecutorService.shutdown();
        executorService.awaitTermination(2, TimeUnit.SECONDS);
        scheduledExecutorService.awaitTermination(2, TimeUnit.SECONDS);
    }
}
